@@ -1,4 +1,4 @@
-; Copyright (c) 1993-2001 by Richard Kelsey and Jonathan Rees. See file COPYING.
+; Copyright (c) 1993-2005 by Richard Kelsey and Jonathan Rees. See file COPYING.
 
 ; Command levels for the command processor
 ;
@@ -38,7 +38,7 @@
 				; lazily generated list of this level's threads
 
 (define (make-command-level condition inspecting? dynamic-env levels throw)
-  (let ((level (really-make-command-level (make-thread-queue)
+  (let ((level (really-make-command-level (make-queue)
 					  (make-counter)
 					  dynamic-env
 					  levels
@@ -52,7 +52,9 @@
 					      #f)
 					  #f	; paused thread
 					  #f)))	; undetermined thread list
-    (spawn-repl-thread! level)
+    (if (user-session-script-mode? (user-session))
+	(spawn-script-thread! level)
+	(spawn-repl-thread! level))
     level))
 
 ; Add THUNK as a thread to LEVEL.  The level is stored in the thread so
@@ -60,13 +62,17 @@
 ; run queue.
 
 (define (spawn-on-command-level level thunk id)    
-  (let ((thread (make-thread thunk (command-level-dynamic-env level) id)))
-    (set-thread-scheduler! thread (command-thread))
-    (set-thread-data! thread level)
-    (enqueue-thread! (command-level-queue level) thread)
-    (increment-counter! (command-level-thread-counter level))
+  (let ((thread (make-thread thunk id)))
+    (spawn-thread-on-command-level level thread)
     thread))
 
+(define (spawn-thread-on-command-level level thread)
+  (set-thread-dynamic-env! thread (command-level-dynamic-env level))
+  (set-thread-scheduler! thread (command-thread))
+  (set-thread-data! thread level)
+  (enqueue! (command-level-queue level) thread)
+  (increment-counter! (command-level-thread-counter level)))
+
 ; Add a new REPL thread to LEVEL.
 
 (define (spawn-repl-thread! level)
@@ -75,6 +81,16 @@
 					'command-loop)))
     (set-command-level-repl-thread! level thread)))
 
+; Add a script thread to LEVEL
+
+(define (spawn-script-thread! level)
+  (spawn-on-command-level level
+			  (let ((thunk
+				 (user-session-script-thunk (user-session))))
+			    (lambda ()
+			      (set-exit-status! (thunk))))
+			  'script))
+
 ; Find all of the threads belonging to LEVEL.  This may be expensive to call
 ; and may not return the correct value if LEVEL is currently running.
 
@@ -104,7 +120,8 @@
 ; session, runs an initial thunk and then pushes a command level.
 
 (define (start-command-levels resume-args context
-			      start-thunk repl-thunk
+			      greeting-thunk start-thunk
+			      repl-thunk script-thunk
 			      condition inspector-state
 			      input-port output-port error-port)
   ;(debug-message "[Starting levels]")
@@ -112,19 +129,24 @@
   (let ((dynamic-env (get-dynamic-env))
 	(session (make-user-session (current-thread)
 				    (or context (make-user-context))
-				    repl-thunk
+				    repl-thunk script-thunk
 				    input-port
 				    output-port
 				    error-port
 				    resume-args	; focus values
 				    #f		; exit status
 				    (and (pair? resume-args)
-					 (equal? (car resume-args) "batch")))))
+					 (equal? (car resume-args) "batch"))
+				    (and (pair? resume-args)
+					 (equal? (car resume-args) "run-script")))))
     (with-handler command-levels-condition-handler
       (lambda ()
 	(let-fluids $command-level-thread? #t
 		    $user-session session
 	  (lambda ()
+	    (if (not (or (user-session-batch-mode? session)
+			 (user-session-script-mode? session)))
+		(greeting-thunk))
 	    ;(debug-message "[start-thunk]")
 	    (start-thunk)
 	    (let ((thunk (really-push-command-level condition
@@ -182,17 +204,18 @@
 ; go to the 
 
 (define (command-levels-condition-handler c next-handler)
-  (cond ((or (warning? c)
-	     (note? c))
-	 (force-output (current-output-port))   ; keep synchronous
-	 (display-condition c (current-error-port))
-	 (unspecific))				; proceed
-        ((error? c)
-	 (force-output (current-output-port))	; keep synchronous
-	 (display-condition c (current-error-port))
-	 (scheme-exit-now 1))
-        (else                           
-         (next-handler))))
+  (let ((c (coerce-to-condition c)))
+    (cond ((or (warning? c)
+	       (note? c))
+	   (force-output (current-output-port))	; keep synchronous
+	   (display-condition c (current-error-port))
+	   (unspecific))		; proceed
+	  ((or (error? c) (bug? c))
+	   (force-output (current-output-port))	; keep synchronous
+	   (display-condition c (current-error-port))
+	   (scheme-exit-now 1))
+	  (else
+	   (next-handler)))))
 
 ;----------------------------------------------------------------
 ; Grab the current continuation, then make a command level and run it.
@@ -244,16 +267,10 @@
 
 (define (terminate-level level)
   (let ((threads (command-level-threads level))
-	(queue (command-level-queue level))
 	(*out?* #f))
     (for-each (lambda (thread)
 		(if (thread-continuation thread)
-		    (begin
-                      (remove-thread-from-queue! thread)
-		      (interrupt-thread thread
-					(lambda ignore
-					  (terminate-current-thread)))
-		      (enqueue-thread! queue thread))))
+		    (terminate-level-thread thread level)))
 	      threads)
     (dynamic-wind
      (lambda ()
@@ -267,16 +284,29 @@
 	 (if (not (null? levels))
 	     (reset-command-input! (car levels))))))))
 
+; Put the thread on the runnable queue if it is not already there and then
+; terminate it.  Termination removes the thread from any blocking queues
+; and interrupts with a throw that will run any pending dynamic-winds.
+
+(define (terminate-level-thread thread level)
+  (let ((queue (command-level-queue level)))
+    (if (not (on-queue? queue thread))
+	(enqueue! queue thread))
+    (terminate-thread! thread)))
+
 (define (reset-command-input! level)
   (let ((repl (command-level-repl-thread level)))
     (if repl
 	(interrupt-thread repl
 			  (lambda return-values
-			    (signal 'reset-command-input)
+			    (signal the-reset-command-input-condition)
 			    (apply values return-values))))))
 
-(define-condition-type 'reset-command-input '())
-(define reset-command-input? (condition-predicate 'reset-command-input))
+(define-condition-type &reset-command-input &condition
+  reset-command-input?)
+
+(define the-reset-command-input-condition
+  (condition (&reset-command-input)))
 
 ; Make sure the input and output ports are available and then run the threads
 ; on LEVEL's queue.
@@ -308,7 +338,7 @@
     (lambda (event args)
       (enum-case event-type event
 	((spawned)
-	 (spawn-on-command-level level (car args) (cadr args))
+	 (spawn-thread-on-command-level level (car args))
 	 #t)
 	((runnable)
 	 (let* ((thread (car args))
@@ -317,8 +347,7 @@
 		  (error "non-command-level thread restarted on a command level"
 			 thread))
 		 ((memq level levels)
-		  (enqueue-thread! (command-level-queue level)
-				   thread))
+		  (enqueue! (command-level-queue level) thread))
 		 (else
 		  (warn "dropping thread from exited command level"
 			thread)))
@@ -351,7 +380,8 @@
 (define (command-level-wait level terminating?)
   (lambda ()
     (cond ((< 0 (counter-value (command-level-thread-counter level)))
-	   (wait))
+	   (wait)
+	   #t)
 	  (terminating?
 	   #f)
 	  ((exit-status)
@@ -377,7 +407,8 @@
   (lambda (thread token args)
     (if (upcall? token)
 	(apply (upcall-procedure token) args)
-	(propogate-upcall thread token args))))
+	(begin
+	  (propogate-upcall thread token args)))))
 
 (define-record-type upcall :upcall
   (make-upcall procedure id)
@@ -460,7 +491,7 @@
     (if repl-thread
 	(begin 
 	  (set-command-level-repl-thread! level #f)
-	  (kill-thread! repl-thread)))))
+	  (terminate-thread! repl-thread)))))
 
 (define-upcall (push-command-level-upcall condition inspecting?
 					  thread dynamic-env)
@@ -522,6 +553,6 @@
 	(begin
 	  (if (eq? paused (command-level-repl-thread level))
 	      (spawn-repl-thread! level))
-	  (interrupt-thread paused terminate-current-thread)
+	  (terminate-thread! paused)	; it's already running, so no enqueue
 	  (set-command-level-paused-thread! level #f))
 	(warn "level has no paused thread" level))))
